from .predictor import NcutPredictor
from .vision_predictor import NcutVisionPredictor

from .dino_predictor import (
    NcutDinov3Predictor,
    NcutDinoPredictor,
    NcutDinoPredictorFeatUp,
    NcutDinoPredictorSR,
)


import torch
from typing import Tuple

from ncut_pytorch.predictor.vision_predictor import NcutVisionPredictor

from ncut_pytorch.predictor.dino import hires_dino_256, hires_dino_512, hires_dino_1024
from ncut_pytorch.predictor.dino import LowResDINO
from ncut_pytorch.predictor.dino.transform import get_input_transform

from ncut_pytorch.predictor.dino.dinov3 import Dinov3Backbone

SUPER_RESOLUTION_MODELS = {
    256: hires_dino_256,
    512: hires_dino_512,
    1024: hires_dino_1024,
}

class NcutDinoPredictorSR(NcutVisionPredictor):
    def __init__(self,
                 input_size: int = 512,
                 dtype: torch.dtype = torch.float16,
                 batch_size: int = 1):
        model, transform = SUPER_RESOLUTION_MODELS[input_size](dtype=dtype)
        super().__init__(model, transform, batch_size)


class NcutDinoPredictorFeatUp(NcutVisionPredictor):
    def __init__(self,
                 input_size: Tuple[int, int] = (512, 512),
                 batch_size: int = 8):
        model = torch.hub.load("huzeyann/FeatUp", 'dino', use_norm=False)
        transform = get_input_transform(resize=input_size)
        super().__init__(model, transform, batch_size)


class NcutDinoPredictor(NcutVisionPredictor):
    def __init__(self,
                 input_size: Tuple[int, int] = (512, 512),
                 dtype: torch.dtype = torch.float32,
                 batch_size: int = 8):
        model = LowResDINO(dtype=dtype)
        transform = get_input_transform(resize=input_size)
        super().__init__(model, transform, batch_size)


class NcutDinov3Predictor(NcutVisionPredictor):
    def __init__(self,
                 input_size: Tuple[int, int] = (2048, 2048),
                 model_cfg: str = "dinov3_vitl16",
                 batch_size: int = 8,
    ):
        model = Dinov3Backbone(model_cfg)
        transform = get_input_transform(resize=input_size)
        super().__init__(model, transform, batch_size)


__all__ = ["farthest_point_sampling"]

import warnings

import fpsample
import numpy as np
import torch

from .device import auto_device
from .math import pca_lowrank

__IS_FAST_FPS_AVAILABLE = False
try:
    from fpsample import bucket_fps_kdline_sampling
    __IS_FAST_FPS_AVAILABLE = True
except ImportError:
    warnings.warn("""package `fpsample` is not installed with latest version. 
    Using a slower implementation (~100ms slower).
    To install fpsample, run:
    >>> pip install fpsample==0.3.3

    if the above pip install fails, please try install build dependencies (Rust) first:
    >>> sudo apt-get update
    >>> sudo apt-get install build-essential cargo rustc -y
    or 
    >>> conda install rust -c conda-forge""")


@torch.no_grad()
def farthest_point_sampling(
    X: torch.Tensor,
    n_sample: int,
    max_draw_ratio: float = 4.0,
    device: str = None,
):
    # if num_data is too large, use random sampling to reduce the load of farthest point sampling

    num_data = X.shape[0]

    num_draw = int(n_sample * max_draw_ratio)
    if num_draw > num_data:
        return _farthest_point_sampling(X, n_sample, device=device)

    # random draw num_draw samples to reduce the load of farthest point sampling
    draw_indices = np.random.permutation(num_data)[:num_draw]
    sampled_indices = _farthest_point_sampling(
        X[draw_indices],
        n_sample=n_sample,
        device=device,
    )
    return draw_indices[sampled_indices]


@torch.no_grad()
def _farthest_point_sampling(
    X: torch.Tensor,
    n_sample: int,
    h: int = 7,
    device: str = None,
):
    num_data = X.shape[0]
    if n_sample >= num_data:
        return np.arange(num_data)

    if isinstance(X, np.ndarray):
        X = torch.from_numpy(X)
    
    device = auto_device(X.device, device)
    X = X.to(device)

    # PCA to reduce the dimension, because fpsample.bucket_fps_kdline_sampling only supports up to 8 dimensions
    if X.shape[1] > 8:
        X = pca_lowrank(X, q=8)

    assert X.ndim == 2, "X should be a 2D tensor"
    assert X.shape[0] > 0, "X should have at least 1 data point"
    assert X.shape[1] <= 8, "fpsample only supports up to 8 dimensions"
    assert X.shape[1] > 0, "X should have at least 1 dimension"
    assert not torch.any(torch.isnan(X)), "X contains NaN"
    assert not torch.any(torch.isinf(X)), "X contains Inf"
    
    if __IS_FAST_FPS_AVAILABLE:
        h = min(h, int(np.log2(num_data)))
        samples_idx = fpsample.bucket_fps_kdline_sampling(
            X.cpu().numpy(), n_sample, h
        )
    else:
        samples_idx = fpsample.fps_npdu_kdtree_sampling(
            X.cpu().numpy(), n_sample,
        )
    samples_idx = samples_idx.astype(np.int64)
    return samples_idx




from typing import List, Tuple, Union

import numpy as np
import torch

from ncut_pytorch import kway_ncut, ncut_fn
from ncut_pytorch import mspace_color, tsne_color, umap_color
from ncut_pytorch.ncuts.ncut_click import ncut_click_prompt
from ncut_pytorch.ncuts.ncut_kway import axis_align
from ncut_pytorch.ncuts.ncut_nystrom import nystrom_propagate
from ncut_pytorch.utils.sample import farthest_point_sampling
from ncut_pytorch.utils.math import chunked_matmul


class NotInitializedError(Exception):
    """Raised when trying to use predictor methods before calling initialize()"""
    pass

class NcutPredictor:
    _initialized: bool = False
    device: str = 'cpu'
    color_method: str = 'umap'

    def __init__(self):
        self._features: torch.Tensor
        self._hierarchy_assign: List[torch.Tensor]
        self._eigvecs: torch.Tensor
        self._color_palette: torch.Tensor

        # inference states
        self._nystrom_indices: torch.Tensor
        self._gamma: float
        self._click_eigvecs: torch.Tensor
        self._R: torch.Tensor
        self._fg_idx: int
        self._bg_idx: int
        
        # kway ncut states
        self._kway_sample_idx: torch.Tensor

    def initialize(self,
                   features: torch.Tensor,
                   n_segments: Union[List[int], int] = (5, 25, 50, 100, 250)
                   ) -> None:
        self._features = features
        if isinstance(n_segments, int):
            n_segments = [n_segments]
        self.refresh_eigvecs(max(n_segments))
        self._initialized = True
        self.cache_hierarchy(n_segments)
        self._color_palette = []

    def refresh_eigvecs(self, n_eig: int) -> None:
        eigvecs, eigval = ncut_fn(self._features, n_eig=n_eig, device=self.device)
        self._eigvecs = eigvecs
        self._kway_sample_idx = farthest_point_sampling(eigvecs, 10240, device=self.device)

    def get_n_eigvecs(self, n_eig: int) -> torch.Tensor:
        cache_hit = n_eig <= self._eigvecs.shape[1]
        if not cache_hit:
            self.refresh_eigvecs(n_eig)
        return self._eigvecs[:, :n_eig]

    def cache_hierarchy(self, n_segments: List[int]) -> None:
        hierarchy_assign = []
        for n_eig in n_segments:
            hierarchy_assign.append(self.get_n_segments(n_eig))
        self._hierarchy_assign = hierarchy_assign

    def get_n_segments(self, n_cluster: int) -> torch.Tensor:
        self.__check_initialized()
        eigvecs = self.get_n_eigvecs(n_cluster)
        kway_eigvec = kway_ncut(eigvecs, device=self.device, sample_idx=self._kway_sample_idx)
        cluster_assignment = kway_eigvec.argmax(dim=1).cpu()
        return cluster_assignment

    def get_hierarchy_masks(self, point_index: int) -> List[torch.Tensor]:
        self.__check_initialized()
        masks: List[torch.Tensor] = []
        for cluster_assignment in self._hierarchy_assign:
            cluster_idx = cluster_assignment[point_index].item()
            mask = cluster_assignment == cluster_idx
            masks.append(mask)
        return masks

    def predict_clicks(self,
                       fg_indices: np.ndarray,
                       bg_indices: np.ndarray,
                       click_weight: float,
                       **kwargs
                       ) -> Tuple[torch.Tensor, torch.Tensor]:
        self.__check_initialized()
        eigvecs, eigval, nystrom_indices, gamma = ncut_click_prompt(
            self._features,
            fg_indices,
            bg_indices,
            return_indices_and_gamma=True,
            click_weight=click_weight,
            **kwargs,
        )

        eigvecs = kway_ncut(eigvecs, device=self.device)
        R = axis_align(eigvecs, device=self.device)
        kway_eigvecs = chunked_matmul(eigvecs, R, device=self.device, large_device=eigvecs.device)

        # find which cluster is the foreground
        fg_eigvecs = kway_eigvecs[fg_indices]
        fg_idx = fg_eigvecs.mean(0).argmax().item()
        bg_idx = 1 if fg_idx == 0 else 0

        # discretize the eigvecs
        mask = kway_eigvecs.argmax(dim=-1) == fg_idx
        heatmap = kway_eigvecs[:, fg_idx] - kway_eigvecs[:, bg_idx]

        # save for inference use
        self._nystrom_indices = nystrom_indices
        self._gamma = gamma
        self._click_eigvecs = eigvecs
        self._R = R
        self._fg_idx = fg_idx
        self._bg_idx = bg_idx

        return mask, heatmap

    def inference_new_features(self, new_features: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        self.__check_initialized()
        if not hasattr(self, "_nystrom_indices") or len(self._nystrom_indices) == 0:
            raise NotInitializedError("please call predict_clicks() before inference_new_features()")

        nystrom_X = self._features[self._nystrom_indices]
        nystrom_out = self._click_eigvecs[self._nystrom_indices]
        eigvecs = nystrom_propagate(nystrom_out, new_features, nystrom_X,
                                    gamma=self._gamma, device=self.device)
        eigvecs = chunked_matmul(eigvecs, self._R, device=self.device, large_device=eigvecs.device)
        mask = eigvecs.argmax(dim=-1) == self._fg_idx
        heatmap = eigvecs[:, self._fg_idx] - eigvecs[:, self._bg_idx]
        return mask, heatmap

    def get_color_palette(self, n_eig: int = 50) -> torch.Tensor:
        cache_hit = hasattr(self, '_color_palette') and len(self._color_palette) > 0
        if not cache_hit:
            self.refresh_color_palette(n_eig)
        return self._color_palette

    def refresh_color_palette(self, n_eig: int = 50) -> None:
        self.__check_initialized()
        if self.color_method == 'mspace':
            self._color_palette = mspace_color(self._eigvecs[:, :n_eig])
        elif self.color_method == 'tsne':
            self._color_palette = tsne_color(self._eigvecs[:, :n_eig])
        elif self.color_method == 'umap':
            self._color_palette = umap_color(self._eigvecs[:, :n_eig])
        else:
            raise ValueError(f"Invalid color method: {self.color_method}")
        
    def inference_new_color_palette(self, new_image: torch.Tensor) -> torch.Tensor:
        ... # TODO: implement this

    def __check_initialized(self) -> None:
        if not self._initialized or not hasattr(self, '_features') or \
            not hasattr(self, '_eigvecs'):
            raise NotInitializedError("Not initialized, please call initialize() first")

    def to(self, device: Union[str, torch.device]):
        self.device = device
        return self



__all__ = ['ncut_click_prompt']

from typing import Callable, Union

import numpy as np
import torch

from ncut_pytorch.utils.gamma import find_gamma_by_degree_after_fps
from ncut_pytorch.utils.math import rbf_affinity, normalize_affinity
from ncut_pytorch.utils.sample import farthest_point_sampling
from ncut_pytorch.utils.device import auto_device
from .ncut_nystrom import NystromConfig
from .ncut_nystrom import nystrom_propagate
from .ncut_nystrom import _plain_ncut


#TODO: automatically optimize click_weight based on the iou of fg and bg
def ncut_click_prompt(
        X: torch.Tensor,
        fg_indices: np.ndarray,
        bg_indices: np.ndarray = None,
        click_weight: float = 0.5,
        bg_weight: float = 0.1,
        n_eig: int = 2,
        track_grad: bool = False,
        d_gamma: float = None,
        device: str = None,
        gamma: float = None,
        affinity_fn: Callable[[torch.Tensor, torch.Tensor, float], torch.Tensor] = rbf_affinity,
        no_propagation: bool = False,
        return_indices_and_gamma: bool = False,
        **kwargs,
) -> Union[tuple[torch.Tensor, torch.Tensor], tuple[torch.Tensor, torch.Tensor, torch.Tensor, float]]:

    config = NystromConfig()
    config.update(kwargs)

    # use GPU if available
    device = auto_device(X.device, device)

    # skip pytorch gradient computation if track_grad is False
    prev_grad_state = torch.is_grad_enabled()
    torch.set_grad_enabled(track_grad)
    
    if bg_indices is None:
        bg_indices = np.array([], dtype=np.int64)

    # subsample for nystrom approximation
    nystrom_indices = farthest_point_sampling(X, n_sample=config.n_sample, device=device)
    nystrom_indices = torch.tensor(nystrom_indices, dtype=torch.long)
    # remove fg and bg from fps_idx
    nystrom_indices = nystrom_indices[~np.isin(nystrom_indices, np.concatenate([fg_indices, bg_indices]))]
    # add fg and bg to fps_idx
    nystrom_indices = np.concatenate([fg_indices, bg_indices, nystrom_indices])
    fg_indices = np.arange(len(fg_indices))
    bg_indices = np.arange(len(bg_indices)) + len(fg_indices)
    n_fgbg = len(fg_indices) + len(bg_indices)
    
    nystrom_X = X[nystrom_indices].to(device)
    
    # find optimal gamma for affinity matrix
    if gamma is None:
        gamma = find_gamma_by_degree_after_fps(nystrom_X, d_gamma, affinity_fn)

    # compute Ncut on the nystrom sampled subgraph
    A = affinity_fn(nystrom_X, gamma=gamma)
    A = normalize_affinity(A)

    # modify the affinity from the clicks
    X_click = 1 * A[fg_indices].mean(0)
    if len(bg_indices) > 0:
        X_click = X_click - bg_weight * A[bg_indices].mean(0)
    
    X_click = X_click * A.shape[0]

    A_click = affinity_fn(X_click.unsqueeze(1), gamma=0.5)
    A_click = normalize_affinity(A_click)
    
    _A = click_weight * A_click + (1 - click_weight) * A
        
    nystrom_eigvec, eigval = _plain_ncut(_A, n_eig)
    
    if no_propagation:
        torch.set_grad_enabled(prev_grad_state)
        return nystrom_eigvec, eigval, nystrom_indices, gamma

    # propagate eigenvectors from subgraph to full graph
    eigvec, nystrom_indices2 = nystrom_propagate(
        nystrom_eigvec,
        X,
        nystrom_X,
        n_neighbors=config.n_neighbors,
        n_sample=config.n_sample2,
        gamma=gamma,
        matmul_chunk_size=config.matmul_chunk_size,
        device=device,
        move_output_to_cpu=config.move_output_to_cpu,
        track_grad=track_grad,
        return_indices=True,
        affinity_fn=affinity_fn,
    )
    
    torch.set_grad_enabled(prev_grad_state)    

    if return_indices_and_gamma:
        indices = nystrom_indices[nystrom_indices2]
        return eigvec, eigval, indices, gamma

    return eigvec, eigval



__all__ = ["mspace_color", "tsne_color", "umap_color", "umap_sphere_color", "rotate_rgb_cube", "convert_to_lab_color"]

import warnings
from typing import Any, Callable, Dict, Literal, Tuple, Optional

import numpy as np
import torch

from .mspace import mspace_viz_transform
from ncut_pytorch.ncuts.ncut_nystrom import nystrom_propagate
from ncut_pytorch.utils.math import quantile_normalize
from ncut_pytorch.utils.sample import farthest_point_sampling


def _identity(X: torch.Tensor) -> torch.Tensor:
    return X


def mspace_color(
        X: torch.Tensor,
        q: float = 0.95,
        n_eig: Optional[int] = 32,
        n_dim: int = 3,
        training_steps: int = 100,
        progress_bar: bool = False,
        **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """

    low_dim_embedding = mspace_viz_transform(
        X=X,
        n_eig=n_eig,
        mood_dim=n_dim,
        training_steps=training_steps,
        progress_bar=progress_bar,
        eigvec_loss=1.0,
        recon_loss=1.0,
        decoder_training_steps=0,
        boundary_loss=0.0,
        zero_center_loss=0.01,
        repulsion_loss=0.01,
        attraction_loss=0,
        axis_align_loss=0.0,
        degree=['auto'],
        **kwargs)

    rgb = rgb_from_nd_colormap(low_dim_embedding, q=q)

    return rgb


def tsne_color(
        X: torch.Tensor,
        num_sample: int = 1000,
        perplexity: int = 150,
        n_dim: int = 3,
        metric: Literal["cosine", "euclidean"] = "cosine",
        device: str = None,
        seed: int = 0,
        q: float = 0.95,
        knn: int = 10,
        **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from sklearn.manifold import TSNE
    except ImportError:
        raise ImportError(
            "sklearn import failed, please install `pip install scikit-learn`"
        )
    num_sample = min(num_sample, X.shape[0])
    if perplexity > num_sample // 2:
        warnings.warn(
            f"perplexity is larger than num_sample, set perplexity to {num_sample // 2}",
            stacklevel=2,
            category=UserWarning,
        )
        perplexity = num_sample // 2

    low_dim_embedding, rgb = _nystrom_dimension_reduction(
        X=X,
        num_sample=num_sample,
        metric=metric,
        rgb_func=rgb_from_nd_colormap,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=TSNE, reduction_dim=n_dim, reduction_kwargs={
            "perplexity": perplexity,
        },
    )

    return rgb


def umap_color(
        X: torch.Tensor,
        num_sample: int = 1000,
        n_neighbors: int = 150,
        min_dist: float = 1.0,
        spread: float = 10.0,
        n_dim: int = 3,
        metric: Literal["cosine", "euclidean"] = "cosine",
        device: str = None,
        seed: int = 0,
        q: float = 0.95,
        knn: int = 10,
        **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from umap import UMAP
    except ImportError:
        raise ImportError("umap import failed, please install `pip install umap-learn`")

    low_dim_embedding, rgb = _nystrom_dimension_reduction(
        X=X,
        num_sample=num_sample,
        metric=metric,
        rgb_func=rgb_from_nd_colormap,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=UMAP, reduction_dim=n_dim, reduction_kwargs={
            "n_neighbors": n_neighbors,
            "min_dist": min_dist,
            "spread": spread,
            "low_memory": False,
        },
    )

    return rgb


def umap_sphere_color(
        X: torch.Tensor,
        num_sample: int = 1000,
        n_neighbors: int = 150,
        min_dist: float = 0.1,
        metric: Literal["cosine", "euclidean"] = "cosine",
        device: str = None,
        seed: int = 0,
        q: float = 0.95,
        knn: int = 10,
        **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from umap import UMAP
    except ImportError:
        raise ImportError("umap import failed, please install `pip install umap-learn`")

    def transform_func(X: torch.Tensor) -> torch.Tensor:
        return torch.stack((
            torch.sin(X[:, 0]) * torch.cos(X[:, 1]),
            torch.sin(X[:, 0]) * torch.sin(X[:, 1]),
            torch.cos(X[:, 0]),
        ), dim=1)

    low_dim_embedding, rgb = _nystrom_dimension_reduction(
        X=X,
        num_sample=num_sample,
        metric=metric,
        rgb_func=rgb_from_nd_colormap,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=UMAP, reduction_dim=2, reduction_kwargs={
            "n_neighbors": n_neighbors,
            "min_dist": min_dist,
            "output_metric": "haversine",
        },
        transform_func=transform_func
    )

    return rgb


def _nystrom_dimension_reduction(
        X: torch.Tensor,
        num_sample: int,
        metric: Literal["cosine", "euclidean"],
        rgb_func: Callable[[torch.Tensor, float], torch.Tensor],
        q: float,
        knn: int,
        seed: int,
        device: str,
        reduction: Callable[..., "sklearn.base.BaseEstimator"],
        reduction_dim: int,
        reduction_kwargs: Dict[str, Any],
        transform_func: Callable[[torch.Tensor], torch.Tensor] = _identity,
) -> Tuple[torch.Tensor, torch.Tensor]:
    subgraph_indices = farthest_point_sampling(X, n_sample=num_sample, device=device)

    _inp = X[subgraph_indices].cpu().numpy()
    _subgraph_embed = reduction(
        n_components=reduction_dim,
        metric=metric,
        random_state=seed,
        **reduction_kwargs
    ).fit_transform(_inp)

    _subgraph_embed = torch.tensor(_subgraph_embed, dtype=torch.float32)
    X_nd = transform_func(nystrom_propagate(
        _subgraph_embed,
        X,
        X[subgraph_indices],
        n_neighbors=knn,
        device=device,
        move_output_to_cpu=True,
    ))
    rgb = rgb_func(X_nd, q)
    return X_nd, rgb


def flatten_sphere(X_3d):
    x = np.arctan2(X_3d[:, 0], X_3d[:, 1])
    y = -np.arccos(X_3d[:, 2])
    X_2d = np.stack([x, y], axis=1)
    return X_2d


def rotate_rgb_cube(rgb, position=1):
    """rotate RGB cube to different position

    Args:
        rgb (torch.Tensor): RGB color space [0, 1], shape (*, 3)
        position (int): position to rotate, 0, 1, 2, 3, 4, 5, 6

    Returns:
        torch.Tensor: RGB color space, shape (n_samples, 3)
    """
    assert position in range(0, 7), "position should be 0, 1, 2, 3, 4, 5, 6"
    rotation_matrix = torch.tensor(
        [
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 0],
        ]
    ).float()
    n_mul = position % 3
    rotation_matrix = torch.matrix_power(rotation_matrix, n_mul)
    rgb = rgb @ rotation_matrix
    if position > 3:
        rgb = 1 - rgb
    return rgb


def rgb_from_3d_rgb_cube(X_3d, q=0.95):
    """convert 3D t-SNE to RGB color space

    Args:
        X_3d (torch.Tensor): 3D t-SNE embedding, shape (n_samples, 3)
        q (float): quantile, default 0.95

    Returns:
        torch.Tensor: RGB color space, shape (n_samples, 3)
    """
    assert X_3d.shape[1] == 3, "input should be (n_samples, 3)"
    assert len(X_3d.shape) == 2, "input should be (n_samples, 3)"
    rgb = []
    for i in range(3):
        rgb.append(quantile_normalize(X_3d[:, i], q=q))
    rgb = torch.stack(rgb, dim=-1)
    return rgb


def rgb_from_2d_colormap(X_2d, q=0.95):
    xy = X_2d.clone()
    for i in range(2):
        xy[:, i] = quantile_normalize(xy[:, i], q=q)

    try:
        from pycolormap_2d import (
            ColorMap2DBremm,
            ColorMap2DZiegler,
            ColorMap2DCubeDiagonal,
            ColorMap2DSchumann,
        )
    except ImportError:
        raise ImportError(
            "pycolormap_2d import failed, please install `pip install pycolormap-2d`"
        )

    cmap = ColorMap2DCubeDiagonal()
    xy = xy.cpu().numpy()
    len_x, len_y = cmap._cmap_data.shape[:2]
    x = (xy[:, 0] * (len_x - 1)).astype(np.int64)
    y = (xy[:, 1] * (len_y - 1)).astype(np.int64)
    x = np.clip(x, 0, len_x - 1)
    y = np.clip(y, 0, len_y - 1)
    rgb = cmap._cmap_data[x, y]
    rgb = torch.tensor(rgb, dtype=torch.float32) / 255
    return rgb


def rgb_from_nd_colormap(X_nd, q=0.95):
    """
    Returns:
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    d = X_nd.shape[1]
    if d == 2:
        return rgb_from_2d_colormap(X_nd, q=q)
    elif d == 3:
        return rgb_from_3d_rgb_cube(X_nd, q=q)
    else:
        raise ValueError(f"Unsupported dimensionality: {d}")


def convert_to_lab_color(rgb, full_range=True):
    from skimage import color
    import copy

    if isinstance(rgb, torch.Tensor):
        rgb = rgb.cpu().numpy()
    _rgb = copy.deepcopy(rgb)
    _rgb[..., 0] = _rgb[..., 0] * 100
    if full_range:
        _rgb[..., 1] = _rgb[..., 1] * 255 - 128
        _rgb[..., 2] = _rgb[..., 2] * 255 - 128
    else:
        _rgb[..., 1] = _rgb[..., 1] * 100 - 50
        _rgb[..., 2] = _rgb[..., 2] * 100 - 50
    lab_rgb = color.lab2rgb(_rgb)
    return lab_rgb
